import numpy as np
import scipy.constants as ct
from wake_t.particles.interpolation import (
    gather_field_cyl_linear,
    gather_main_fields_cyl_linear,
)
from wake_t.utilities.bunch_generation import get_matched_bunch


def test_gather_field_cyl_linear():
    """
    Test the method `gather_field_cyl_linear`.

    A 2D array is generated and then gathered exactly at the position
    of the field values. If everything works perfectly, both the original
    and the gathered array should be identical.

    """
    # Create field (r: cell centered, z: node centered)
    n_r = 1000
    n_z = 2000
    r_min = 0
    r_max = 10
    z_min = -10
    z_max = 10
    dr = (r_max - r_min) / n_r
    dz = (z_max - z_min) / (n_z - 1)
    r = np.linspace(dr / 2, r_max - dr / 2, n_r)
    z = np.linspace(z_min, z_max, n_z)
    R, Z = np.meshgrid(r, z)
    f = np.zeros((n_z + 4, n_r + 4))
    f[2:-2, 2:-2] = np.sin(R) * np.sin(Z)

    # Particles positioned exactly at grid nodes.
    x_part = R.flatten()
    y_part = np.zeros_like(x_part)
    z_part = Z.flatten()

    # Gather field
    f_part = gather_field_cyl_linear(
        f, z_min, z_max, r_min + dr / 2, r_max, dz, dr, x_part, y_part, z_part
    )

    # Check
    f_part = np.reshape(f_part, (n_z, n_r))
    np.testing.assert_array_almost_equal(f_part, f[2:-2, 2:-2])


def test_gather_main_fields_cyl_linear():
    """
    Test the method `gather_main_fields_cyl_linear`.

    Same test idea as `test_gather_field_cyl_linear`. A 2D array is generated
    which is given as the value of the Wr and Ez fields. These fields are then
    gathered at the exact location of the field values. If everything works
    perfectly, both the original and the gathered arrays should be identical.

    """
    # Create field (r: cell centered, z: node centered)
    n_r = 1000
    n_z = 2000
    r_min = 0
    r_max = 10
    z_min = -10
    z_max = 10
    dr = (r_max - r_min) / n_r
    dz = (z_max - z_min) / (n_z - 1)
    r = np.linspace(dr / 2, r_max - dr / 2, n_r)
    z = np.linspace(z_min, z_max, n_z)
    R, Z = np.meshgrid(r, z)
    f = np.zeros((n_z + 4, n_r + 4))
    f[2:-2, 2:-2] = np.sin(R) * np.sin(Z)

    # Particles positioned exactly at grid nodes.
    x_part = R.flatten()
    y_part = np.zeros_like(x_part)
    z_part = Z.flatten()

    # Preallocate field arrays
    n_part = x_part.shape[0]
    ex = np.zeros(n_part)
    ey = np.zeros(n_part)
    ez = np.zeros(n_part)
    bx = np.zeros(n_part)
    by = np.zeros(n_part)
    bz = np.zeros(n_part)

    # Gather field
    gather_main_fields_cyl_linear(
        f,
        f,
        f,
        z_min,
        z_max,
        r_min + dr / 2,
        r_max,
        dz,
        dr,
        x_part,
        y_part,
        z_part,
        ex,
        ey,
        ez,
        bx,
        by,
        bz,
    )

    ex = np.reshape(ex, (n_z, n_r))
    ey = np.reshape(ey, (n_z, n_r))
    ez = np.reshape(ez, (n_z, n_r))
    bx = np.reshape(bx, (n_z, n_r))
    by = np.reshape(by, (n_z, n_r))
    bz = np.reshape(bz, (n_z, n_r))

    # Check
    np.testing.assert_array_almost_equal(ex, f[2:-2, 2:-2])
    np.testing.assert_array_almost_equal(ez, f[2:-2, 2:-2])
    np.testing.assert_array_almost_equal(by, f[2:-2, 2:-2])


def test_gather_main_fields_cyl_linear_at_bunch():
    """
    Test `gather_main_fields_cyl_linear` by gathering fields at the location
    of the particles of a bunch. This test checks that the sum of the
    gathered field values has not changed.

    """
    # Bunch parameters.
    en = 1e-6  # m
    ene = 200  # units of beta*gamma
    ene_sp = 0.3  # %
    xi_c = 0.0  # m
    s_t = 3  # fs
    q_tot = 1  # pC
    n_part = 1e4

    # Set numpy random seed to get reproducible results
    np.random.seed(1)

    # Generate bunch (in this case, matched to a density of 10^23 cm-3)
    bunch = get_matched_bunch(en, en, ene, ene_sp, s_t, xi_c, q_tot, n_part, n_p=1e23)

    # Create field (r: cell centered, z: node centered)
    n_r = 1000
    n_z = 2000
    r_min = 0
    r_max = 30e-6
    z_min = -10e-6
    z_max = 10e-6
    dr = (r_max - r_min) / n_r
    dz = (z_max - z_min) / (n_z - 1)
    r = np.linspace(dr / 2, r_max - dr / 2, n_r)
    z = np.linspace(z_min, z_max, n_z)
    R, Z = np.meshgrid(r, z)
    f = np.zeros((n_z + 4, n_r + 4))
    f[2:-2, 2:-2] = np.sin(R / r_max - 1) * np.sin(Z / z_max - 1)

    # Get preallocated field arrays
    ex, ey, ez, bx, by, bz = bunch.get_field_arrays()

    # Gather field
    gather_main_fields_cyl_linear(
        f,
        f,
        f,
        z_min,
        z_max,
        r_min + dr / 2,
        r_max,
        dz,
        dr,
        bunch.x,
        bunch.y,
        bunch.xi,
        ex,
        ey,
        ez,
        bx,
        by,
        bz,
    )

    # Check
    np.testing.assert_almost_equal(np.sum(ex / ct.c + by), 15.709780373078333)
    np.testing.assert_almost_equal(np.sum(ey / ct.c - bx), -75.57014245377374)
    np.testing.assert_almost_equal(np.sum(ez), 6798.124201568496)


if __name__ == "__main__":
    test_gather_field_cyl_linear()
    test_gather_main_fields_cyl_linear()
    test_gather_main_fields_cyl_linear_at_bunch()



from pytest import approx
import numpy as np

from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_twiss
from wake_t.beamline_elements.active_plasma_lens import ActivePlasmaLens
from wake_t.diagnostics import analyze_bunch


def test_active_plasma_lens():
    """
    Check that the beam propagation through an active plasma lens without
    wakefields is accurate.
    """
    # Set numpy random seed to get reproducible results
    np.random.seed(1)

    # generate bunch
    en = 1e-6
    alpha = 0
    beta = 1.0
    s_t = 10
    ene = 600
    ene_sp = 2
    bunch = get_gaussian_bunch_from_twiss(
        en_x=en,
        en_y=en,
        a_x=alpha,
        a_y=alpha,
        b_x=beta,
        b_y=beta,
        ene=ene,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0e-6,
        q_tot=10,
        n_part=1e5,
    )
    apl = ActivePlasmaLens(1e-2, 1000, wakefields=False, n_out=3)

    apl.track(bunch)
    bunch_params = analyze_bunch(bunch)
    gamma_x = bunch_params["gamma_x"]
    assert approx(gamma_x, rel=1e-10) == 92.38407651552583


def test_active_plasma_lens_with_wakefields():
    """
    Check that the beam propagation through an active plasma lens with
    wakefields is accurate.
    """
    # Set numpy random seed to get reproducible results.
    np.random.seed(1)

    # Generate bunch.
    en = 1e-6
    alpha = 0
    beta = 1.0
    s_t = 10
    ene = 600
    ene_sp = 2
    bunch = get_gaussian_bunch_from_twiss(
        en_x=en,
        en_y=en,
        a_x=alpha,
        a_y=alpha,
        b_x=beta,
        b_y=beta,
        ene=ene,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0e-6,
        q_tot=10,
        n_part=1e5,
    )

    # Define APL.
    apl = ActivePlasmaLens(
        1e-2,
        1000,
        wakefields=True,
        n_out=3,
        density=1e23,
        wakefield_model="quasistatic_2d",
        r_max=200e-6,
        xi_min=-30e-6,
        xi_max=30e-6,
        n_r=200,
        n_xi=120,
        dz_fields=0.2e-3,
        ppc=4,
    )

    # Do tracking.
    apl.track(bunch)

    # Analyze and check results.
    bunch_params = analyze_bunch(bunch)
    gamma_x = bunch_params["gamma_x"]
    assert approx(gamma_x, rel=1e-10) == 77.43788529521473


if __name__ == "__main__":
    test_active_plasma_lens()
    test_active_plasma_lens_with_wakefields()



from typing import Literal, Optional

import numpy as np
import scipy.constants as ct

from .plasma_stage import DtBunchType
from .field_element import FieldElement
from wake_t.physics_models.em_fields.quadrupole import QuadrupoleField


class FieldQuadrupole(FieldElement):
    """
    Class defining a quadrupole as a field element.

    Parameters
    ----------
    length : float
        Length of the quadrupole lens in :math:`m`.
    foc_strength : float
        Focusing strength of the quadrupole in :math:`T/m`. Defined so
        that a positive value is focusing for electrons in the :math:`x`
        plane and defocusing in the :math:`y` plane.
    dt_bunch : float, str, or list of float and str
        The time step for evolving the particle bunches. If ``'auto'``, it will
        be automatically set to :math:`dt = T/(10*2*pi)`, where T is the
        betatron period of the particle with the lowest energy in the bunch.
        A list of values can also be provided. In this case, the list
        should have the same order as the list of bunches given to the
        ``track`` method.
    bunch_pusher : str
        The pusher used to evolve the particle bunches in time within
        the specified fields. Possible values are ``'rk4'`` (Runge-Kutta
        method of 4th order) or ``'boris'`` (Boris method).
    n_out : int, optional
        Number of times along the lens in which the particle distribution
        should be returned (A list with all output bunches is returned
        after tracking).
    name : str, optional
        Name of the quadrupole. This is only used for displaying the
        progress bar during tracking. By default, 'quadrupole'
    """

    def __init__(
        self,
        length: float,
        foc_strength: float,
        dt_bunch: Optional[DtBunchType] = "auto",
        bunch_pusher: Literal["boris", "rk4"] = "boris",
        n_out: Optional[int] = 1,
        name: Optional[str] = "quadrupole",
    ) -> None:
        self.foc_strength = foc_strength
        super().__init__(
            length=length,
            dt_bunch=dt_bunch,
            bunch_pusher=bunch_pusher,
            n_out=n_out,
            name=name,
            fields=[QuadrupoleField(foc_strength)],
            auto_dt_bunch=self._get_optimized_dt,
        )

    def _get_optimized_dt(self, beam):
        """Get tracking time step."""
        # Get minimum gamma in the bunch (assumes px,py << pz).
        q_over_m = beam.q_species / beam.m_species
        min_gamma = np.sqrt(np.min(beam.pz) ** 2 + 1)
        w_x = np.sqrt(np.abs(q_over_m * ct.c * self.foc_strength / min_gamma))
        T_x = 1 / w_x
        dt = 0.1 * T_x
        return dt
